﻿/*-------------------------------------------------------------------------
 Cipher algorithm interface
 
 File name   : $FileName: Program.cs $
 Version     : $Rev$
 Created     : Jan/14/2011
 Last update : $Date$
 Desgined by : $Author$
 
 Copyright (C) 2011 AIST.
 
 By using this code, you agree to the following terms and conditions.
 
 This code is copyrighted by AIST and Tohoku University ("us").
 
 Permission is hereby granted to copy, reproduce, redistribute or
 otherwise use this code as long as: there is no monetary profit gained
 specifically from the use or reproduction of this code, it is not sold,
 rented, traded or otherwise marketed, and this copyright notice is
 included prominently in any copy made.
 
 We shall not be liable for any damages, including without limitation
 direct, indirect, incidental, special or consequential damages arising
 from the use of this code.
 
 When you publish any results arising from the use of this code, we will
 appreciate it if you can cite our webpage
 (http://www.rcis.aist.go.jp/special/SASEBO/).
 -------------------------------------------------------------------------*/
using System;
using System.Collections.Generic;
using System.Linq;

namespace RoundKey
{
  static class RK
  {
    public static readonly string VERSION;
    private static int[] RCon;

    static RK()
    {
      RCon = new int[] {
        0x8d, 0x01, 0x02, 0x04, 0x08, 0x10,
        0x20, 0x40, 0x80, 0x1b, 0x36, 0x6c
      };
      System.Reflection.Assembly assembly = System.Reflection.Assembly.GetExecutingAssembly();
      VERSION = "Ver." + assembly.GetName().Version.ToString();
    }

    private static int rotWord(int a)
    {
      return a >> 8 & 0x00ffffff | (a & 0xff) << 24;
    }

    private static int subWord(int a)
    {
      return AES.Sbox[a & 0xff] & 0xff
           | (AES.Sbox[a >> 8 & 0xff] & 0xff) << 8
           | (AES.Sbox[a >> 16 & 0xff] & 0xff) << 16
           | (AES.Sbox[a >> 24 & 0xff] & 0xff) << 24;
    }

    private static int[,] expandsKey(byte[] key)
    {
      int nk = key.Length / 4;
      int nr = nk + 6;
      int[,] rk = new int[nr + 1, 4];

      for (int i = 0; i < nk; i++)
      {
        rk[i >> 2, i & 3] = key[i * 4] & 0xff
                          | (key[i * 4 + 1] & 0xff) << 8
                          | (key[i * 4 + 2] & 0xff) << 16
                          | (key[i * 4 + 3] & 0xff) << 24;
      }

      int x = rk[nk - 1 >> 2, nk - 1 & 3];
      for (int i= nk; i < (nr + 1) * 4; i++)
      {
        if (i % nk == 0)
        {
          x = subWord(rotWord(x)) ^ RCon[i / nk];
        }
        if (nk == 8 && i % nk == 4)
        {
          x = subWord(x);
        }
        x ^= rk[i - nk >> 2, i - nk & 3];
        rk[i >> 2, i & 3] = x;
      }
      return rk;
    }

    private static int[,] invExpandKey(byte[] key)
    {
      int nk = key.Length / 4;
      int nr = nk + 6;
      int[,] invRk = new int[nr + 1, 4];

      int x;

      for (int i = 0; i < nk; i++)
      {
        invRk[nr, i & 3] = (key[i * 4] & 0xff)
                         | (key[i * 4 + 1] & 0xff) << 8
                         | (key[i * 4 + 2] & 0xff) << 16
                         | (key[i * 4 + 3] & 0xff) << 24;
      }

      for (int i = (nr + 1) * 4 - 1; i >= nk; i--)
      {
        int j = i - 1;
        if ((i - nk) % nk == 0)
        {
          x = subWord(rotWord(invRk[j >> 2, j & 3])) ^ RCon[i / nk] ^ invRk[i >> 2, i & 3];
        }
        else
        {
          x = invRk[j >> 2, j & 3] ^ invRk[i >> 2, i & 3];
        }
        invRk[i - nk >> 2, i - nk & 3] = x;
      }
      return invRk;
    }

    private static void Usage()
    {
      Console.Error.WriteLine(@"Usage: RoundKey.EXE [-r] key");
      Console.Error.WriteLine(@"Build round keys (of AES-128) from given key.");
      Console.Error.WriteLine(@"Key must be separated with SPACE like this:");
      Console.Error.WriteLine("");
      Console.Error.WriteLine(@" RoundKey.EXE ""00 01 02 03 04 05 06 07 08 09 0a 0b 0c 0d 0e 0f""");
      Console.Error.WriteLine("");
      Console.Error.WriteLine(@" -r indicates to reverse the key as 10th round key to obtain initial key.");
      Console.Error.WriteLine("");
      Console.Error.WriteLine(@" Copyright (C) 2011 AIST {0}", VERSION);
      return;
    }

    static void Main(string[] args)
    {
      try
      {
        List<byte> l = new List<byte>();
        int[,] result;

        if (args.Length == 1)
        {
          foreach (var b in args[0].Split(new char[] { }, StringSplitOptions.RemoveEmptyEntries))
          {
            l.Add(Convert.ToByte(b, 16));
          }
          result = expandsKey(l.ToArray<Byte>());
        }
        else if (args.Length > 1)
        {
          foreach (var b in args[1].Split(new char[] { }, StringSplitOptions.RemoveEmptyEntries))
          {
            l.Add(Convert.ToByte(b, 16));
          }
          result = invExpandKey(l.ToArray<Byte>());
        }
        else
        {
          Usage();
          return;
        }

        for (int i = 0; i <= 10; i++)
        {
          Console.Write("{0:D02}: ", i);
          for (int j = 0; j < result.GetLength(1); j++)
          {
            int tmp = result[i, j];
            Console.Write(String.Format("{0:X02} ", (tmp) & 0xff));
            Console.Write(String.Format("{0:X02} ", (tmp >> 8) & 0xff));
            Console.Write(String.Format("{0:X02} ", (tmp >> 16) & 0xff));
            Console.Write(String.Format("{0:X02} ", (tmp >> 24) & 0xff));
          }
          Console.WriteLine();
        }
#if DEBUG
        while (true)
        {
          System.Threading.Thread.Sleep(1000);
        }
#endif
      }
      catch (Exception ex)
      {
        Console.WriteLine("Oops! something wrong.");
        Console.WriteLine(ex.Message);
      }
    }
  }
}
